import logging
import mimetypes
from datetime import date
from io import BytesIO
from typing import List

from drc_cmis.client.mapper import (
    CONNECTION_MAP,
    DOCUMENT_MAP,
    GEBRUIKSRECHTEN_MAP,
    OBJECTINFORMATIEOBJECT_MAP,
    REVERSE_CONNECTION_MAP,
    REVERSE_DOCUMENT_MAP,
    mapper,
)
from drc_cmis.client.utils import get_random_string

from .utils import CMISRequest

logger = logging.getLogger(__name__)


class CMISBaseObject(CMISRequest):
    def __init__(self, data):
        super().__init__()
        self.data = data
        self.properties = dict(data.get("properties", {}))

    def reload(self):
        logger.debug("CMIS: DRC_DOCUMENT: reload")
        params = {
            "objectId": self.objectId,
            "cmisselector": "object",
        }

        json_response = self.get_request(self.root_folder_url, params=params)
        self.data = json_response
        self.properties = json_response.get("properties")

    def get_object_parents(self):
        logger.debug("CMIS: DRC_DOCUMENT: get_object_parents")
        params = {
            "objectId": self.objectId,
            "cmisselector": "parents",
        }

        json_response = self.get_request(self.root_folder_url, params=params)
        return self.get_all_objects(json_response, Folder)

    def move(self, sourceFolder, targetFolder):
        data = {
            "objectId": self.objectId,
            "cmisaction": "move",
            "sourceFolderId": sourceFolder.objectId,
            "targetFolderId": targetFolder.objectId,
        }
        logger.debug(f"From: {sourceFolder.name} To: {targetFolder.name}")
        logger.debug(
            f"From: {sourceFolder.objectTypeId} To: {targetFolder.objectTypeId}"
        )
        # invoke the URL
        json_response = self.post_request(self.root_folder_url, data=data)
        self.data = json_response
        self.properties = json_response.get("properties")
        return self


class Document(CMISBaseObject):
    table = "drc:document"
    object_type_id = f"D:{table}"

    def __getattr__(self, name: str):
        convert_string = f"drc:{name}"
        if name in DOCUMENT_MAP:
            convert_string = DOCUMENT_MAP.get(name)
        elif name in CONNECTION_MAP:
            convert_string = CONNECTION_MAP.get(name)
        elif (
            convert_string not in REVERSE_CONNECTION_MAP
            and convert_string not in REVERSE_DOCUMENT_MAP
        ):
            convert_string = f"cmis:{name}"

        if convert_string not in self.properties:
            raise AttributeError(f"No property '{convert_string}'")

        return self.properties[convert_string]["value"]

    @classmethod
    def build_properties(
        cls, data: dict, new: bool = True, identification: str = ""
    ) -> dict:
        logger.debug(
            "Building CMIS properties, document identification: %s",
            identification or "(not set)",
        )

        props = {}
        for key, value in data.items():
            prop_name = mapper(key, type="document")
            if not prop_name:
                logger.debug("No property name found for key '%s'", key)
                continue
            props[prop_name] = value

        if new:
            props.setdefault("cmis:objectTypeId", cls.object_type_id)

            # increase likelihood of uniqueness of title by appending a random string
            title, suffix = data.get("titel"), get_random_string()
            if title is not None:
                props["cmis:name"] = f"{title}-{suffix}"

            # make sure the identification is set, but _only_ for newly created documents.
            # identificatie is immutable once the document is created
            if identification:
                prop_name = mapper("identificatie")
                props[prop_name] = identification

        # can't or shouldn't be written
        props.pop(mapper("uuid"), None)

        return props

    def checkout(self):
        logger.debug("CMIS: DRC_DOCUMENT: checkout")
        data = {"objectId": self.objectId, "cmisaction": "checkOut"}
        json_response = self.post_request(self.root_folder_url, data=data)
        return Document(json_response)

    def update_properties(self, properties):
        logger.debug("CMIS: DRC_DOCUMENT: update_properties")
        data = {
            "objectId": self.objectId,
            "cmisaction": "update",
        }
        prop_count = 0
        for prop_key, prop_value in properties.items():
            # Skip property because update is not allowed
            if prop_key == "cmis:objectTypeId":
                continue

            if isinstance(prop_value, date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%I:%S.000Z")

            data["propertyId[%s]" % prop_count] = prop_key
            data["propertyValue[%s]" % prop_count] = prop_value
            prop_count += 1

        # invoke the URL
        json_response = self.post_request(self.root_folder_url, data=data)
        self.data = json_response
        self.properties = json_response.get("properties")

        return self

    def get_private_working_copy(self) -> "Document":
        """
        Retrieve the private working copy version of a document.
        """
        logger.debug("Fetching the PWC for the current document (UUID '%s')", self.uuid)
        # http://docs.oasis-open.org/cmis/CMIS/v1.1/os/CMIS-v1.1-os.html#x1-5590004
        params = {
            "cmisselector": "object",  # get the object rather than the content
            "objectId": self.versionSeriesCheckedOutId,
        }
        data = self.get_request(self.root_folder_url, params)
        return type(self)(data)

    def checkin(self, checkin_comment, major=True):
        logger.debug("CMIS: DRC_DOCUMENT: checkin")
        props = {
            "objectId": self.objectId,
            "cmisaction": "checkIn",
            "checkinComment": checkin_comment,
            "major": major,
        }

        # invoke the URL
        json_response = self.post_request(self.root_folder_url, props)
        return Document(json_response)

    def set_content_stream(self, content_file):
        logger.debug("CMIS: DRC_DOCUMENT: set_content_stream")
        data = {
            "objectId": self.objectId,
            "cmisaction": "setContent",
        }

        mimetype = None
        # need to determine the mime type
        if not mimetype and hasattr(content_file, "name"):
            mimetype, _encoding = mimetypes.guess_type(content_file.name)

        if not mimetype:
            mimetype = "application/binary"

        files = {self.name: (self.name, content_file, mimetype)}

        json_response = self.post_request(self.root_folder_url, data=data, files=files)
        return Document(json_response)

    def get_content_stream(self) -> BytesIO:
        logger.debug("CMIS: DRC_DOCUMENT: get_content_stream")
        params = {
            "objectId": self.objectId,
            "cmisaction": "content",
        }
        file_content = self.get_request(self.root_folder_url, params=params)
        return BytesIO(file_content)

    @classmethod
    def get_all_versions(cls, document: "Document") -> List["Document"]:
        """
        Retrieve all versions for a given document.

        Versions are ordered by most-recent first based on cmis:creationDate. If there
        is a PWC, it shall be the first object.

        http://docs.oasis-open.org/cmis/CMIS/v1.1/errata01/os/CMIS-v1.1-errata01-os-complete.html#x1-3440006
        """

        params = {"objectId": document.objectId, "cmisselector": "versions"}
        all_versions = document.get_request(document.root_folder_url, params=params)
        return [cls(data) for data in all_versions]

    def destroy(self) -> None:
        """
        Permanently delete the object from the CMIS store, with all its versions.

        By default, all versions should be deleted according to the CMIS standard. If
        the document is currently locked (i.e. there is a private working copy), we need
        to cancel that checkout first.
        """
        if self.isVersionSeriesCheckedOut:
            pwc = self.get_private_working_copy()
            cancel_checkout_data = {
                "cmisaction": "cancelCheckout",
                "objectId": pwc.objectId,
            }
            self.post_request(self.root_folder_url, data=cancel_checkout_data)

        data = {"objectId": self.objectId, "cmisaction": "delete"}
        self.post_request(self.root_folder_url, data=data)


class Gebruiksrechten(CMISBaseObject):
    table = "drc:gebruiksrechten"

    def __getattr__(self, name):
        """
        :param name: Name of the attribute to retrieve
        :type name: string
        :return: the attribute (Note: date times are returned as timestamps)
        """
        convert_string = f"drc:{name}"
        if name in GEBRUIKSRECHTEN_MAP:
            # Properties characteristic only of this Gebruiksrechten type
            convert_string = GEBRUIKSRECHTEN_MAP.get(name)
        else:
            # General alresco properties
            convert_string = f"cmis:{name}"
        return self.properties.get(convert_string, {}).get("value")

    def delete_gebruiksrechten(self):
        data = {"objectId": self.objectId, "cmisaction": "delete"}
        json_response = self.post_request(self.root_folder_url, data=data)
        return json_response


class ObjectInformatieObject(CMISBaseObject):
    table = "drc:oio"

    def __getattr__(self, name):
        """
        :param name: Name of the attribute to retrieve
        :type name: string
        :return: the attribute (Note: date times are returned as timestamps)
        """
        convert_string = f"drc:{name}"
        if name in OBJECTINFORMATIEOBJECT_MAP:
            # Properties characteristic only of this ObjectInformatieObject type
            convert_string = OBJECTINFORMATIEOBJECT_MAP.get(name)
        else:
            # General alresco properties
            convert_string = f"cmis:{name}"
        return self.properties.get(convert_string, {}).get("value")

    def delete_oio(self):
        data = {"objectId": self.objectId, "cmisaction": "delete"}
        json_response = self.post_request(self.root_folder_url, data=data)
        return json_response


class Folder(CMISBaseObject):
    def __getattr__(self, name):
        return self.properties.get(f"cmis:{name}", {}).get("value")

    def get_children(self):
        logger.debug("CMIS: DRC_DOCUMENT: get_children")
        data = {
            "cmisaction": "query",
            "statement": f"SELECT * FROM cmis:folder WHERE IN_FOLDER('{self.objectId}')",
        }
        json_response = self.post_request(self.base_url, data=data)
        return self.get_all_results(json_response, Folder)

    def create_folder(self, name, properties={}, **kwargs):
        logger.debug("CMIS: DRC_DOCUMENT: create_folder")
        data = {
            "objectId": self.objectId,
            "cmisaction": "createFolder",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": name,
        }

        data["propertyId[1]"] = "cmis:objectTypeId"
        if "cmis:objectTypeId" in properties.keys():
            data["propertyValue[1]"] = properties.pop("cmis:objectTypeId")
        else:
            data["propertyValue[1]"] = "cmis:folder"

        prop_count = 2
        for prop, value in properties.items():
            data[f"propertyId[{prop_count}]"] = prop
            data[f"propertyValue[{prop_count}]"] = value
            prop_count += 1

        json_response = self.post_request(self.root_folder_url, data=data)
        return Folder(json_response)

    def create_document(self, name, properties, content_file=None, **kwargs):
        logger.debug("CMIS: DRC_DOCUMENT: create_document")
        data = {
            "objectId": self.objectId,
            "cmisaction": "createDocument",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": name,
        }

        data["propertyId[1]"] = "cmis:objectTypeId"
        if "cmis:objectTypeId" in properties.keys():
            data["propertyValue[1]"] = properties.pop("cmis:objectTypeId")
        else:
            data["propertyValue[1]"] = "drc:document"

        prop_count = 2
        for prop_key, prop_value in properties.items():
            if isinstance(prop_value, date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%M:%S.000Z")

            data[f"propertyId[{prop_count}]"] = prop_key
            data[f"propertyValue[{prop_count}]"] = prop_value
            prop_count += 1

        json_response = self.post_request(self.root_folder_url, data=data)
        cmis_doc = Document(json_response)
        return cmis_doc.set_content_stream(content_file)

    def create_gebruiksrechten(self, name, properties):
        data = {
            "objectId": self.objectId,
            "cmisaction": "createDocument",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": name,
        }

        data["propertyId[1]"] = "cmis:objectTypeId"
        if "cmis:objectTypeId" in properties.keys():
            data["propertyValue[1]"] = properties.pop("cmis:objectTypeId")
        else:
            data["propertyValue[1]"] = "D:drc:gebruiksrechten"

        prop_count = 2
        for prop_key, prop_value in properties.items():
            if isinstance(prop_value, date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%M:%S.000Z")

            data[f"propertyId[{prop_count}]"] = prop_key
            data[f"propertyValue[{prop_count}]"] = prop_value
            prop_count += 1

        json_response = self.post_request(self.root_folder_url, data=data)
        return Gebruiksrechten(json_response)

    def create_oio(self, name, properties):
        data = {
            "objectId": self.objectId,
            "cmisaction": "createDocument",
            "propertyId[0]": "cmis:name",
            "propertyValue[0]": name,
        }

        data["propertyId[1]"] = "cmis:objectTypeId"
        if "cmis:objectTypeId" in properties.keys():
            data["propertyValue[1]"] = properties.pop("cmis:objectTypeId")
        else:
            data["propertyValue[1]"] = "D:drc:oio"

        prop_count = 2
        for prop_key, prop_value in properties.items():
            if isinstance(prop_value, date):
                prop_value = prop_value.strftime("%Y-%m-%dT%H:%M:%S.000Z")

            data[f"propertyId[{prop_count}]"] = prop_key
            data[f"propertyValue[{prop_count}]"] = prop_value
            prop_count += 1

        json_response = self.post_request(self.root_folder_url, data=data)
        return ObjectInformatieObject(json_response)

    def delete_tree(self, **kwargs):
        data = {"objectId": self.objectId, "cmisaction": "deleteTree"}
        self.post_request(self.root_folder_url, data=data)



import logging
from datetime import datetime
from decimal import Decimal
from io import BytesIO
from typing import List, Optional, Union
from uuid import UUID

from django.db.utils import IntegrityError
from django.utils.crypto import constant_time_compare

from cmislib.exceptions import UpdateConflictException

from drc_cmis.cmis.drc_document import (
    Document,
    Folder,
    Gebruiksrechten,
    ObjectInformatieObject,
)
from drc_cmis.cmis.utils import CMISRequest

from .exceptions import (
    DocumentConflictException,
    DocumentDoesNotExistError,
    DocumentExistsError,
    DocumentLockConflictException,
    DocumentNotLockedException,
    GetFirstException,
)
from .mapper import mapper
from .query import CMISQuery
from .utils import get_random_string

logger = logging.getLogger(__name__)


class CMISDRCClient(CMISRequest):
    """
    DRC client implementation using the CMIS protocol.

    This class makes it easier to query for everything you need.
    """

    documents_in_folders_query = CMISQuery("SELECT * FROM drc:document WHERE %s")
    find_folder_by_case_url_query = CMISQuery(
        "SELECT * FROM drc:zaakfolder WHERE drc:zaak__url='%s'"
    )

    _repo = None
    _root_folder = None
    _base_folder = None

    @property
    def _get_base_folder(self):
        logger.debug("CMIS_CLIENT: _get_base_folder")
        if not self._base_folder:
            base = self.get_request(self.root_folder_url)
            for folder_response in base.get("objects"):
                folder = Folder(folder_response["object"])
                if folder.name == self.base_folder:
                    self._base_folder = folder
            if not self._base_folder:
                folder = Folder({})
                self._base_folder = folder.create_folder(name=self.base_folder)
        return self._base_folder

    # generic querying
    def query(self, return_type, lhs: List[str], rhs: List[str]):
        table = return_type.table
        where = (" WHERE " + " AND ".join(lhs)) if lhs else ""
        query = CMISQuery("SELECT * FROM %s%s" % (table, where))

        body = {
            "cmisaction": "query",
            "statement": query(*rhs),
        }
        response = self.post_request(self.base_url, body)
        logger.debug(response)
        return self.get_all_results(response, return_type)

    # ZRC Notification client calls.
    def get_or_create_zaaktype_folder(self, zaaktype):
        """
        Create a folder with the prefix 'zaaktype-' to make a zaaktype folder

        """
        logger.debug("CMIS_CLIENT: get_or_create_zaaktype_folder")
        properties = {
            "cmis:objectTypeId": "F:drc:zaaktypefolder",
            mapper("url", "zaaktype"): zaaktype.get("url"),
            mapper("identificatie", "zaaktype"): zaaktype.get("identificatie"),
        }

        folder_name = (
            f"zaaktype-{zaaktype.get('omschrijving')}-{zaaktype.get('identificatie')}"
        )
        cmis_folder = self._get_or_create_folder(
            folder_name, self._get_base_folder, properties
        )
        return cmis_folder

    def get_or_create_zaak_folder(self, zaak, zaaktype_folder):
        """
        Create a folder with the prefix 'zaak-' to make a zaak folder

        """
        logger.debug("CMIS_CLIENT: get_or_create_zaak_folder")
        properties = {
            "cmis:objectTypeId": "F:drc:zaakfolder",
            mapper("url", "zaak"): zaak.get("url"),
            mapper("identificatie", "zaak"): zaak.get("identificatie"),
            mapper("zaaktype", "zaak"): zaak.get("zaaktype"),
            mapper("bronorganisatie", "zaak"): zaak.get("bronorganisatie"),
        }
        cmis_folder = self._get_or_create_folder(
            f"zaak-{zaak.get('identificatie')}", zaaktype_folder, properties
        )
        return cmis_folder

    # DRC client calls.
    def create_document(self, identification: str, data: dict, content=None):
        """
        Create a cmis document.

        Args:
            identification (str): A unique identifier for the document.
            data (dict): A dict with all the data that needs to be saved on the document.
            content (BytesStream): The content of the document.

        Returns:
            AtomPubDocument: A cmis document.

        Raises:
            DocumentExistsError: when a document already uses the same identifier.

        """
        logger.debug("CMIS_CLIENT: create_document")
        self._check_document_exists(identification)

        now = datetime.now()
        data.setdefault("versie", 1)

        if content is None:
            content = BytesIO()

        year_folder = self._get_or_create_folder(str(now.year), self._get_base_folder)
        month_folder = self._get_or_create_folder(str(now.month), year_folder)
        day_folder = self._get_or_create_folder(str(now.day), month_folder)

        properties = Document.build_properties(
            data, new=True, identification=identification
        )

        # Make sure that the content is set.
        content.seek(0)
        return day_folder.create_document(
            name=properties.pop("cmis:name"),
            properties=properties,
            content_file=content,
        )

    def get_cmis_documents(self, filters=None, page=1, results_per_page=100):
        """
        Gives a list of cmis documents.

        Args:
            filters (dict): A dict of filters that need to be applied to the search.

        Returns:
            AtomPubDocument: A list of CMIS documents.

        """
        logger.debug("CMIS_CLIENT: get_cmis_documents")
        filter_string = self._build_filter(filters, strip_end=True)
        query = "SELECT * FROM drc:document WHERE drc:document__verwijderd='false'"
        if filter_string:
            query += f" AND {filter_string}"

        logger.debug(query)
        data = {
            "cmisaction": "query",
            "statement": query,
        }

        skip_count = 0
        if page:
            results_per_page = results_per_page
            max_items = results_per_page
            skip_count = page * results_per_page - results_per_page

            data["maxItems"] = max_items
            data["skipCount"] = skip_count

        json_response = self.post_request(self.base_url, data)
        logger.debug(json_response)
        results = self.get_all_results(json_response, Document)
        return {
            "has_next": json_response["hasMoreItems"],
            "total_count": json_response["numItems"],
            "has_prev": skip_count > 0,
            "results": results,
        }

    def get_cmis_document(
        self, uuid: Optional[str], via_identification=None, filters=None
    ):
        """
        Given a cmis document instance.

        :param uuid: UUID of the document as used in the endpoint URL
        :return: :class:`AtomPubDocument` object, the latest version of this document
        """
        logger.debug("CMIS_CLIENT: get_cmis_document")
        assert (
            not via_identification
        ), "Support for 'via_identification' is being dropped"

        error_string = (
            f"Document met identificatie {uuid} bestaat niet in het CMIS connection"
        )
        does_not_exist = DocumentDoesNotExistError(error_string)

        # shortcut - no reason in going over the wire
        if uuid is None:
            raise does_not_exist

        # this always selects the latest version
        query = CMISQuery("SELECT * FROM drc:document WHERE cmis:objectId = '%s' %s")

        filter_string = self._build_filter(
            filters, filter_string="AND ", strip_end=True
        )
        data = {
            "cmisaction": "query",
            "statement": query(uuid, filter_string),
        }
        json_response = self.post_request(self.base_url, data)

        try:
            return self.get_first_result(json_response, Document)
        except GetFirstException as exc:
            raise does_not_exist from exc

    def update_cmis_document(self, uuid: str, lock: str, data: dict, content=None):
        logger.debug("Updating document with UUID %s", uuid)
        cmis_doc = self.get_cmis_document(uuid)

        if not cmis_doc.isVersionSeriesCheckedOut:
            raise DocumentNotLockedException(
                "Document is not checked out and/or locked."
            )

        assert not cmis_doc.isPrivateWorkingCopy, "Unexpected PWC retrieved"
        pwc = cmis_doc.get_private_working_copy()

        if not pwc.lock:
            raise DocumentNotLockedException(
                "Document is not checked out and/or locked."
            )

        correct_lock = constant_time_compare(lock, pwc.lock)
        if not correct_lock:
            raise DocumentLockConflictException("Wrong document lock given.")

        # build up the properties
        current_properties = cmis_doc.properties
        new_properties = Document.build_properties(data, new=False)

        diff_properties = {
            key: value
            for key, value in new_properties.items()
            if current_properties.get(key) != value
        }

        try:
            pwc.update_properties(diff_properties)
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc

        if content is not None:
            pwc.set_content_stream(content)

        return pwc

    def delete_cmis_document(self, uuid):
        """
        Update the property that the document is deleted.

        Arguments:
            uuid (str): The uuid of the document.

        Returns:
            AtomPubDocument: A CMIS document.

        Raises:
            DocumentConflictException: If the document could not be updated.
        """
        logger.debug("CMIS_CLIENT: delete_cmis_document")
        cmis_doc = self.get_cmis_document(uuid)
        new_properties = {mapper("verwijderd"): True}

        try:
            cmis_doc.update_properties(new_properties)
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc
        return cmis_doc

    def delete_cmis_folders_in_base(self):
        for child_folder in self._get_base_folder.get_children():
            child_folder.delete_tree()

    def obliterate_document(self, uuid: str) -> None:
        logger.debug("CMIS_CLIENT: obliterate_document")
        cmis_doc = self.get_cmis_document(uuid)
        cmis_doc.destroy()
        logger.debug("CMIS_CLIENT: obliteration successful")

    # Split ########################################################################################

    def update_case_connection(self, uuid, data):
        logger.debug("CMIS_CLIENT: update_case_connection")
        cmis_doc = self.get_cmis_document(uuid)

        current_properties = cmis_doc.properties
        new_properties = self._build_case_properties(data, allow_empty=False)
        diff_properties = {
            key: value
            for key, value in new_properties.items()
            if current_properties.get(key) != new_properties.get(key)
        }

        if diff_properties:
            try:
                cmis_doc.update_properties(diff_properties)
            except UpdateConflictException as exc:
                # Node locked!
                raise DocumentConflictException from exc
        return cmis_doc

    def delete_case_connection(self, uuid):
        logger.debug("CMIS_CLIENT: delete_case_connection")
        cmis_doc = self.get_cmis_document(uuid)
        parents = cmis_doc.get_object_parents()
        connection_count = len(parents)

        in_zaakfolder = False
        for parent in parents:
            if parent.properties.get("cmis:objectTypeId") == "F:drc:zaakfolder":
                in_zaakfolder = True

        if in_zaakfolder and connection_count == 1:
            now = datetime.now()
            year_folder = self._get_or_create_folder(
                str(now.year), self._get_base_folder
            )
            month_folder = self._get_or_create_folder(str(now.month), year_folder)
            day_folder = self._get_or_create_folder(str(now.day), month_folder)

            self.move_to_case(cmis_doc, day_folder)

        # Clear properties
        current_properties = cmis_doc.properties
        new_properties = self._build_case_properties({}, allow_empty=True)
        diff_properties = {
            key: value
            for key, value in new_properties.items()
            if current_properties.get(key) != new_properties.get(key)
        }

        if diff_properties:
            try:
                cmis_doc.update_properties(diff_properties)
            except UpdateConflictException as exc:
                # Node locked!
                raise DocumentConflictException from exc
        return cmis_doc

    def move_to_case(self, cmis_doc, folder):
        logger.debug("CMIS_CLIENT: move_to_case")
        parent = [parent for parent in cmis_doc.get_object_parents()][0]
        cmis_doc.move(parent, folder)

    def copy_document(self, cmis_doc, folder, data):
        """
        The copy from source is not supported via the atom pub bindings.
        So we create a new document with the same values.
        """
        logger.debug("CMIS_CLIENT: copy_document")

        attributes = [
            "auteur",
            "beschrijving",
            "bestandsnaam",
            "bronorganisatie",
            "creatiedatum",
            "formaat",
            "identificatie",
            "indicatie_gebruiksrecht",
            "informatieobjecttype",
            "integriteit_algoritme",
            "integriteit_datum",
            "integriteit_waarde",
            "link",
            "ondertekening_datum",
            "ondertekening_soort",
            "ontvangstdatum",
            "status",
            "taal",
            "vertrouwelijkheidaanduiding",
            "verwijderd",
            "verzenddatum",
        ]

        # copy the properties from the source document
        properties = {
            mapper(attribute, type="document"): getattr(cmis_doc, attribute)
            for attribute in attributes
        }

        properties.update(
            **{
                "cmis:objectTypeId": cmis_doc.objectTypeId,
                mapper("titel", type="document"): f"{cmis_doc.titel} - copy",
                "drc:kopie_van": cmis_doc.uuid,  # Keep tack of where this is copied from.
            }
        )

        new_properties = self._build_case_properties(data)
        properties.update(**new_properties)

        if not folder:
            now = datetime.now()
            year_folder = self._get_or_create_folder(
                str(now.year), self._get_base_folder
            )
            month_folder = self._get_or_create_folder(str(now.month), year_folder)
            folder = self._get_or_create_folder(str(now.day), month_folder)

        # Update the cmis:name to make it more unique
        file_name = f"{cmis_doc.titel}-{get_random_string()}"
        properties["cmis:name"] = file_name

        stream = cmis_doc.get_content_stream()
        new_doc = folder.create_document(
            name=file_name, properties=properties, content_file=stream,
        )

        return new_doc

    def get_folder_from_case_url(self, zaak_url):
        logger.debug("CMIS_CLIENT: get_folder_from_case_url")
        query = self.find_folder_by_case_url_query(zaak_url)
        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)
        try:
            return self.get_first_result(json_response, Folder)
        except GetFirstException:
            return None

    # Private functions.
    def _get_or_create_folder(self, name, parent, properties=None):
        """
        Get or create the folder with :param:`name` in :param:`parent`.

        :param name: string, the name of the folder to create.
        :param parent: parent folder to create the folder in as subfolder.
        :param properties: dictionary with cmis and/or custom properties to
          pass to the folder object
        :return: the folder that was retrieved or created.
        """
        logger.debug("CMIS_CLIENT: _get_or_create_folder")
        existing = self._get_folder(name, parent)
        if existing:
            return existing
        return parent.create_folder(name, properties=properties or {})

    def _get_folder(self, name, parent):
        logger.debug("CMIS_CLIENT: _get_folder")
        existing = next(
            (child for child in parent.get_children() if str(child.name) == str(name)),
            None,
        )
        if existing is not None:
            return existing
        return None

    def _build_filter(self, filters, filter_string="", strip_end=False):
        logger.debug("CMIS_CLIENT: _build_filter")
        if filters:
            for key, value in filters.items():
                if mapper(key):
                    key = mapper(key)
                elif mapper(key, type="connection"):
                    key = mapper(key, type="connection")
                elif mapper(key, type="gebruiksrechten"):
                    key = mapper(key, type="gebruiksrechten")
                elif mapper(key, type="objectinformatieobject"):
                    key = mapper(key, type="objectinformatieobject")

                if value and value in ["NULL", "NOT NULL"]:
                    filter_string += f"{key} IS {value} AND "
                elif isinstance(value, Decimal):
                    filter_string += f"{key} = {value} AND "
                elif isinstance(value, list):
                    if len(value) == 0:
                        continue
                    filter_string += "( "
                    for item in value:
                        sub_filter_string = self._build_filter(
                            {key: item}, strip_end=True
                        )
                        filter_string += f"{sub_filter_string} OR "
                    filter_string = filter_string[:-3]
                    filter_string += " ) AND "
                elif value:
                    filter_string += f"{key} = '{value}' AND "

        if strip_end and filter_string[-4:] == "AND ":
            filter_string = filter_string[:-4]

        return filter_string

    def _build_case_properties(self, data, allow_empty=True):
        logger.debug("CMIS_CLIENT: _build_case_properties")
        props = {}
        if data.get("object") or allow_empty:
            props[mapper("object", "connection")] = data.get("object")
        if data.get("object_type") or allow_empty:
            props[mapper("object_type", "connection")] = data.get("object_type")
        if data.get("aard_relatie") or allow_empty:
            props[mapper("aard_relatie", "connection")] = data.get("aard_relatie")
        if data.get("titel") or allow_empty:
            props[mapper("titel", "connection")] = data.get("titel")
        if data.get("beschrijving") or allow_empty:
            props[mapper("beschrijving", "connection")] = data.get("beschrijving")
        if data.get("registratiedatum") or allow_empty:
            props[mapper("registratiedatum", "connection")] = data.get(
                "registratiedatum"
            )

        return props

    def _check_document_exists(self, identification: Union[str, UUID]):
        # FIXME: should be both bronorganisatie and identification check, not just identification
        logger.debug("CMIS_CLIENT: _check_document_exists")

        column = mapper("identificatie", type="document")
        query = CMISQuery(f"SELECT * FROM drc:document WHERE {column} = '%s'")

        data = {
            "cmisaction": "query",
            "statement": query(str(identification)),
        }
        json_response = self.post_request(self.base_url, data)
        if json_response["numItems"] > 0:
            error_string = f"Document identificatie {identification} is niet uniek."
            raise DocumentExistsError(error_string)

    def create_cmis_gebruiksrechten(self, data):
        """
        Creates a Gebruiksrechten object.
        :return:
        """

        gebruiksrechten_folder = self._get_or_create_folder(
            "Gebruiksrechten", self._get_base_folder
        )

        properties = {
            mapper(key, type="gebruiksrechten"): value
            for key, value in data.items()
            if mapper(key, type="gebruiksrechten")
        }

        return gebruiksrechten_folder.create_gebruiksrechten(
            name=get_random_string(), properties=properties
        )

    def get_all_cmis_gebruiksrechten(self):

        query = "SELECT * FROM drc:gebruiksrechten"

        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)
        results = self.get_all_results(json_response, Gebruiksrechten)
        return {
            "has_next": json_response["hasMoreItems"],
            "total_count": json_response["numItems"],
            "has_prev": False,
            "results": results,
        }

    def get_a_cmis_gebruiksrechten(self, uuid):

        query = f"SELECT * FROM drc:gebruiksrechten WHERE cmis:objectId = 'workspace://SpacesStore/{uuid};1.0'"

        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)

        try:
            return self.get_first_result(json_response, Gebruiksrechten)
        except GetFirstException:
            error_string = (
                f"Gebruiksrechten met uuid {uuid} bestaat niet in het CMIS connection"
            )
            raise DocumentDoesNotExistError(error_string)

    def get_cmis_gebruiksrechten(self, filters):

        if filters.get("uuid") is not None:
            results = [self.get_a_cmis_gebruiksrechten(filters.get("uuid"))]
            return {
                "has_next": False,
                "total_count": 1,
                "has_prev": False,
                "results": results,
            }
        else:
            query = "SELECT * FROM drc:gebruiksrechten WHERE "
            sql_filters = self._build_filter(filters, strip_end=True)

            if sql_filters:
                query += f"{sql_filters}"

            data = {
                "cmisaction": "query",
                "statement": query,
            }

            json_response = self.post_request(self.base_url, data)
            results = self.get_all_results(json_response, Gebruiksrechten)
            return {
                "has_next": json_response["hasMoreItems"],
                "total_count": json_response["numItems"],
                "has_prev": False,
                "results": results,
            }

    def delete_cmis_geruiksrechten(self, uuid):

        gebruiksrechten = self.get_a_cmis_gebruiksrechten(uuid)

        try:
            gebruiksrechten.delete_gebruiksrechten()
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc

    def create_cmis_oio(self, data):
        """
        Creates a ObjectInformatieObject.
        """

        # TODO: Implement constraints directly in Alfresco?
        if data.get("zaak") is not None and data.get("besluit") is not None:
            raise IntegrityError(
                "ObjectInformatie object cannot have both Zaak and Besluit relation"
            )
        elif data.get("zaak") is None and data.get("besluit") is None:
            raise IntegrityError(
                "ObjectInformatie object needs to have either a Zaak or Besluit relation"
            )
        oio_folder = self._get_or_create_folder(
            "ObjectInformatieObject", self._get_base_folder
        )

        properties = {
            mapper(key, type="objectinformatieobject"): value
            for key, value in data.items()
            if mapper(key, type="objectinformatieobject")
        }

        return oio_folder.create_oio(name=get_random_string(), properties=properties)

    def get_all_cmis_oio(self):

        query = "SELECT * FROM drc:oio"

        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)
        results = self.get_all_results(json_response, ObjectInformatieObject)
        return {
            "has_next": json_response["hasMoreItems"],
            "total_count": json_response["numItems"],
            "has_prev": False,
            "results": results,
        }

    def get_a_cmis_oio(self, uuid):
        """
        Filters the objectinformatieobject in Alfresco based on the UUID
        :param uuid: string
        :return: The first retrieved objectinformatie object with right uuid
        """

        query = f"SELECT * FROM drc:oio WHERE cmis:objectId = 'workspace://SpacesStore/{uuid};1.0'"

        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)

        try:
            return self.get_first_result(json_response, ObjectInformatieObject)
        except GetFirstException:
            error_string = f"ObjectInformatieObject met uuid {uuid} bestaat niet in het CMIS connection"
            raise DocumentDoesNotExistError(error_string)

    def get_cmis_oio(self, filters):
        """
        Filters the ObjectInformatieObjects on either the informatieobject, the zaak or besluit URL.
        :param filters: dict - valid keys are 'informatieobject', 'zaak' or 'besluit'
        :return: dictionary with the total number of results, the documents retrieved and 'has_prev/next' properties
        """

        if filters.get("uuid") is not None:
            try:
                results = [self.get_a_cmis_oio(filters.get("uuid"))]
                return {
                    "has_next": False,
                    "total_count": 1,
                    "has_prev": False,
                    "results": results,
                }
            except DocumentDoesNotExistError:
                results = []
                return {
                    "has_next": False,
                    "total_count": 0,
                    "has_prev": False,
                    "results": results,
                }
        elif len(filters) == 0:
            query = "SELECT * FROM drc:oio"
        else:
            query = "SELECT * FROM drc:oio WHERE "
            sql_filters = self._build_filter(filters, strip_end=True)

            if sql_filters:
                query += f"{sql_filters}"

        data = {
            "cmisaction": "query",
            "statement": query,
        }

        json_response = self.post_request(self.base_url, data)
        results = self.get_all_results(json_response, ObjectInformatieObject)
        return {
            "has_next": json_response["hasMoreItems"],
            "total_count": json_response["numItems"],
            "has_prev": False,
            "results": results,
        }

    def delete_cmis_oio(self, uuid):
        oio = self.get_a_cmis_oio(uuid)

        try:
            oio.delete_oio()
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc



from django.utils.crypto import get_random_string as _get_random_string


def get_random_string(number: int = 6) -> str:
    allowed_chars = "ABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789"
    return _get_random_string(length=number, allowed_chars=allowed_chars)



from django.db import models

from solo.models import SingletonModel


class CMISConfig(SingletonModel):
    client_url = models.URLField(
        default="http://localhost:8082/alfresco/api/-default-/public/cmis/versions/1.1/browser",
        help_text="API URL for DMS. For example, for alfresco this can be "
        "http://domain_name:port_number/alfresco/api/-default-/public/cmis/versions/1.1/browser",
    )
    client_user = models.CharField(
        max_length=200, default="admin", help_text="Username for logging into DMS",
    )
    client_password = models.CharField(
        max_length=200, default="admin", help_text="Password for logging into DMS"
    )
    base_folder = models.CharField(
        max_length=200,
        default="DRC",
        help_text="Name of the DMS folder in which the documents will be stored.",
    )

    def __str__(self):
        return "CMIS Configuration"

    class Meta:
        verbose_name = "CMIS Configuration"



"""
Listen to the notifications that are send by the NRC
"""
from typing import Type

from django.conf import settings
from django.utils.module_loading import import_string

from vng_api_common.models import APICredential
from vng_api_common.notifications.handlers import default
from zds_client.client import Client

from drc_cmis.client import CMISDRCClient

try:
    from zgw_consumers.client import get_client_class
except ImportError:

    def get_client_class() -> Type[Client]:
        return Client


def get_client(url: str):
    """
    Retrieve a ZDS Client instance for the given API URL.

    The client respects the setting ``CUSTOM_CLIENT_FETCHER`` as used by vng-api-common,
    and will have the auth configured.
    """
    Client = get_client_class()
    default = Client.from_url(url)

    if getattr(settings, "CUSTOM_CLIENT_FETCHER", None):
        client = import_string(settings.CUSTOM_CLIENT_FETCHER)(default.base_url)
        return client
    else:
        default.auth = APICredential.get_auth(url)
        return default


class ZakenHandler:
    def __init__(self):
        self.cmis_client = CMISDRCClient()

    def handle(self, data: dict) -> None:
        if data.get("actie") == "create" and data.get("resource") == "zaak":
            zaaktype_url = data.get("kenmerken", {}).get("zaaktype")
            zaak_url = data.get("resource_url")

            if zaak_url and zaaktype_url:
                zaaktype_folder = self.create_zaaktype_folder(zaaktype_url)
                self.create_zaak_folder(zaak_url, zaaktype_url, zaaktype_folder)

    def create_zaaktype_folder(self, zaaktype_url: str):
        client = get_client(zaaktype_url)
        zaaktype_data = client.retrieve("zaaktype", url=zaaktype_url)
        return self.cmis_client.get_or_create_zaaktype_folder(zaaktype_data)

    def create_zaak_folder(self, zaak_url: str, zaaktype_url: str, zaaktype_folder):
        client = get_client(zaak_url)
        zaak_data = client.retrieve("zaak", url=zaak_url)
        self.cmis_client.get_or_create_zaak_folder(zaak_data, zaaktype_folder)


class RoutingHandler:
    def __init__(self, config: dict, default):
        self.config = config
        self.default = default

    def handle(self, message: dict):
        handler = self.config.get(message["kanaal"])
        if handler is not None:
            handler.handle(message)
        else:
            self.default.handle(message)


default = RoutingHandler({"zaken": ZakenHandler()}, default=default)
